 TTL Fast LOAD routine for 6809 FLEX
*====================================================================
* FILE: FASTLOAD.ASM
* PURPOSE: Faster LOAD routine for TSC 6809 FLEX.
*
* NOTES:
*  - Reads and processes binary load records from disk (physical or
*    virtual) faster than with the built-in FLEX routine.
*  - Installs below MEMEND.
*  - This installer needs to be run each time after FLEX has been booted.
*  - A 20KB program loads from RAM-Disk in approx. 1/2 second @ 1 MHz.
*  - Is fast enough to allow "every-other" interleaving on a 5.25"
*    double-density floppy disk @ 1 MHz.
*  - No improvement for floppy disks if not formatted with optimum
*    interleaving factor.
*  - Original version was written in 1985 and included in RAMDSK29.CMD
*    program for ST-2900 RAM-512 board from Sardis Technologies.
*  - Improved standalone version created 2022-Sep-17.
*
* Last modified 2022-Oct-11 16:52 PDT by David C Wiens.
*====================================================================

*--------------------------------------------------------------------
* Equates.
*--------------------------------------------------------------------
*
* FLEX error codes.
*
NOER EQU 0 no error
NFER EQU 4 no such file found
EFER EQU 8 read past end of file
*
* FLEX System FCB location and FCB field offsets.
*
SYSFCB EQU $C840 System FCB

FES EQU 1 8-bit FCB error status code
FDI EQU 34 8-bit Data Index (offset in sector, 0,4..255)
FSB EQU 64 256-byte Sector Buffer
*
* FMS function codes.
*
CLSF EQU 4 close file
*
* FLEX system variables.
*
LSTTRM EQU $CC11 last terminator
OFFSET EQU $CC1B 16-bit loader address offset
XFR EQU $CC1D transfer flag (non-zero = address found)
TADR EQU $CC1E 16-bit transfer address
ERTYP EQU $CC20 error type
MEMEND EQU $CC2B 16-bit end of user memory
*
* FLEX system routine entry points (in jump tables).
*
WARMS EQU $CD03 warm start
DLOAD EQU $CD30 binary file loader
DRPTER EQU $CD3F report error
FMSCMN EQU $D406 FMS command entry
*
* FLEX binary record type codes.
*
BDES EQU $02 binary load record
TDES EQU $16 transfer address record

 ORG $C100 load installer into Utility Command Area

*--------------------------------------------------------------------
* FUNCTION: INSTALL
* PURPOSE: Install and activate the new LOAD routine.
* NOTES:
*  - Doesn't patch the LOAD vector at $CD30 because FLEX internally
*    calls LOAD directly in several places, bypassing that vector.
*  - Returns to FLEX via jump to WARMS.
* OUT: A,B,X,Y  undefined
*--------------------------------------------------------------------

INSTALL BRA INST2
VN FCB 1 version number
 FCC "cDCW"
*
* Initialize.
*
INST2 PSHS CC
 ORCC #$50 disable interrupts
 NOP
*
* Allocate memory below MEMEND for new FLOAD code.
*
 LDD MEMEND reduce MEMEND by size of new code
 SUBD #FLSIZ
 STD MEMEND
*
* Patch old LOAD code to jump to new FLOAD code.
*
 ADDD #1 installed address of new FLOAD code
 TFR D,Y
 LDX DLOAD+1 address of old LOAD routine
 STD 1,X
 LDA #$7E JMP extended instruction
 STA 0,X
*
* Move new FLOAD code to allocated location.
*
 LDX #FLOAD
IN20 LDA ,X+
 STA ,Y+
 CMPX #FLEND
 BLO IN20
*
* Exit installer.
*
 PULS CC restore interrupts
 JMP WARMS return to FLEX

*--------------------------------------------------------------------
* FUNCTION: FLOAD
* PURPOSE: Fast replacement for FLEX LOAD routine.
* NOTES:
*  - The binary file must already have been opened for binary reading
*    using the System FCB.
*  - OFFSET ($CC1B) must either contain 0 or a loader offset value.
*  - When end-of-file reached and file closed successfully, returns to
*    LOAD caller, but if any error, reports error, then jumps to FLEX
*    WARMS.
* IN: 
* OUT: CC.C  0 = no error
*      A,B,X  undefined
*--------------------------------------------------------------------

FLOAD EQU *
*
* Initialize.
*
 PSHS Y,DP,B,A save Y, reserve 3 bytes working storage

 CLR XFR Transfer Address flag: 0 = not found
 LDX #SYSFCB point to FCB
 LDB FDI,X get current Data Index
 LEAX FSB,X address of Sector Buffer in FCB
*
* Get next byte from file, determine record type.
*
BL10 EQU *

 BSR READB
 CMPA #BDES binary load record?
 BEQ BL30 .Y
 CMPA #TDES transfer address record?
 BNE BL10 .N, invalid code, discard byte
*
* Process transfer address record.
*
 BSR READB
 STA TADR+0 MSB of Transfer Address
 BSR READB
 STA TADR+1 LSB of Transfer Address
 LDA #1 Transfer Address flag: non-zero = found
 STA XFR
 BRA BL10 process next record
*
* Process binary load record header.
*
BL30 BSR READB MSB of load address
 STA 1,S
 BSR READB LSB of load address
 STA 2,S
 BSR READB number of data bytes in load record
 STA 0,S
 BEQ BL10 load record is empty, discard

 PSHS B
 LDD 1+1,S add offset to load address
 ADDD OFFSET
 TFR D,Y
 PULS B

*--------------------------------------------------------------------
* PURPOSE: Process binary load record data payload.
* NOTES:
*  - if data index = 0:  get one byte with READB (uses FMS call) and
*    store to RAM, takes many cycles because also includes reading next
*    sector from disk
*  - if copy block size = 1:  get one byte with READB and store to RAM,
*    takes 65 or 69 cycles
*  - if copy block size = 2..251:  copy block directly from sector
*    buffer to RAM, takes 91 or 95 cycles plus 12.5 cycles/byte
*    (116 or 120 cycles for 2 bytes, 3216 or 3220 cycles for 250
*    bytes)
*  - returns via BRA BL10, but if end-of-file or error, returns
*    directly to LOAD caller or FLEX WARMS
* IN: X  address of sector buffer in FCB
*     Y  address of destination RAM
*     B  current data index (0,5..255)
*     0,S  number of unprocessed bytes in load record
* OUT: Y  incremented by size of data payload
*      B  data index updated (0,5..255)
*      0,S  byte count decremented to zero
*--------------------------------------------------------------------
*
* Calculate size of largest possible copy block, which is
* unprocessed portion of current load record residing in
* unread portion of current sector buffer.
*
BD100 TFR B,A (6) convert data index (0,5..255) into
 NEGA (2) . buffer byte count (0,251..1)
 CMPA 0,S (4) is buffer byte count <= record byte count?
 BLS BD120 (3) .Y, copy block size is buffer byte count
 LDA 0,S (4) .N, copy block size is record byte count

BD120 CMPA #2 (2) is copy block size 2..251 bytes?
 BLO BD170 (3) .N, use single byte copy routine
*
* Various other calculations.
*
 ANDA #$FE (2) round copy block size down to even value

 PSHS X,B,A (9) save X,B, copy block size

 LDA 0+4,S (5) subtract copy blk size frm record byte cnt
 SUBA 0,S (4)
 STA 0+4,S (5)

 CLRA (2) extend 8-bit data index to 16 bits
 LEAX D,X (8) address of source block in sector buffer

 ADDB 0,S (4) increment data index by copy block size
 STB 1,S (5)

 LSR 0,S (6) convert copy block size in bytes to words
*
* Copy specified block from binary load record in sector buffer
* to RAM.  Loop takes 25 cycles per 2 bytes.
* IN: X  address of source block in sector buffer
*     Y  address of destination RAM
*     0,S  size of block to copy (1..125 16-bit words)
*
BD150 LDD ,X++ (8) copy one word
 STD ,Y++ (8)
 DEC 0,S (6) all words in copy block processed?
 BNE BD150 (3) .N, process next word
*
* Clean up stack, determine what to process next.
*
 PULS A,B,X (9) discard word count, restore B,X

 TST 0,S (6) all bytes in binary record processed?
 BNE BD100 (3) .N, process next byte

 BRA BL10 (3) .Y, process next record
*
* Copy single byte from binary record in sector buffer to RAM.
*
BD170 BSR READB (27+) get next data byte from record
 STA ,Y+ (6) store data byte in desired location
 DEC 0,S (6) all data bytes in record processed?
 BNE BD100 (3) .N, process next data byte

 BRA BL10 (3) .Y, process next load record

*--------------------------------------------------------------------
* FUNCTION: READB
* PURPOSE: Read next byte of binary file, either directly from sector
*          buffer or via FMS call.
* NOTES:
*  - Takes 27 cycles (including calling BSR) when reading byte
*    directly from sector buffer in FCB.
*  - If end-of-file or error, returns directly to LOAD caller or to
*    FLEX WARMS.
* IN: X  address of sector buffer in FCB
*     B  current data index (0,5..255)
* OUT: A  data byte
*      B  data index updated (0,5..255)
*--------------------------------------------------------------------

READB CMPB #5 (2) first data byte of sector?
 BLO RB20 (3) .Y, use FMS to get it

 CLRA (2) .N, convert 8-bit offset to 16 bits
 LDA D,X (8) get data byte directly from sector buffer
 INCB (2) increment data index
 RTS (5)

RB20 LEAX -FSB,X point to FCB
 STB FDI,X save current data index
 JSR FMSCMN use FMS call to read data byte
 BNE FMSERR .error
 LDB FDI,X get updated data index
 LEAX FSB,X point to sector buffer
 RTS

*--------------------------------------------------------------------
* PURPOSE: Process FMS error.
* NOTES:
*  - Does not return to READB.
*  - Returns to LOAD caller if end of file and file close successful.
*  - Returns to FLEX WARMS if any error.
* IN: X  address of system FCB
*--------------------------------------------------------------------

FMSERR LEAS 2+3,S discard READB return addr and working storage
 PULS Y restore Y

FE20 LDA FES,X get FMS error code
 CMPA #EFER read past end of file?
 BNE FE70 .N, other error

 LDA #CLSF .Y, function code = close file
 STA 0,X
 JSR FMSCMN use FMS call to close file
 BNE FE20 .error
 ANDCC #$FE .OK, clear CC.C, return to LOAD caller
 RTS

FE70 STA ERTYP save error type code
 JSR DRPTER report error
 CLR LSTTRM
 JMP WARMS abort LOAD and return to FLEX

FLEND EQU *
FLSIZ EQU FLEND-FLOAD

**************************************************************************
*
* Copyright (c) 1985,2022 David C. Wiens, Langley BC, Canada.
*
* Permission to use, copy, modify, and distribute this software for any
* purpose with or without fee is hereby granted, provided that the above
* copyright notice and this permission notice appear in all copies.
*
* THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
* WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
* MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
* ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
* WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
* ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
* OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
*
**************************************************************************

 END INSTALL
